#! /usr/bin/env python
#*******************************************************************************
# ALMA - Atacama Large Millimiter Array
# (c) Associated Universities Inc., 2009 
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
#
# "@(#) $Id: ObsCalAllSkyPointing.py 214236 2015-02-06 05:12:53Z nphillip $"

#
# forcing global imports is due to an OSS problem
#
global CCL
import CCL.Global

global Control
import Control

global ControlExceptionsImpl
import ControlExceptionsImpl

global Observation
import Observation.ObsCalBase
import Observation.AtmCalTarget
import Observation.DelayCalTarget
import Observation.PointingCalTarget
import Observation.FocusCalTarget


class ObsCalAllSkyPointing(Observation.ObsCalBase.ObsCalBase):

    def __init__(self):
        Observation.ObsCalBase.ObsCalBase.__init__(self)
        self._srcPointFocus = None
        self._reverseSpecs = False
        self.WVRResult = None

    def parseOptions(self):
        self.repeatCount             = self.getExpertParameterOrDefault("RepeatCount", int, 1)
        self.pointingSubscanDuration = self.getExpertParameterOrDefault("PointingSubscanDuration", float, 3.456)
        self.atmSubscanDuration      = self.getExpertParameterOrDefault("AtmSubscanDuration", float, 5.76)
        self.sbrSubscanDuration      = self.getExpertParameterOrDefault("SBRSubscanDuration", float, 5.76)
        self.focusSubscanDuration    = self.getExpertParameterOrDefault("FocusSubscanDuration", float, 5.76)
        self.dumpDuration            = self.getExpertParameterOrDefault("dumpDuration", float, 0.192)
        self.channelAverageDuration  = self.getExpertParameterOrDefault("channelAverageDuration", float, 0.576)
        self.integrationDuration     = self.getExpertParameterOrDefault("integrationDuration", float, 0.576)
        #self.atmIntegrationDuration  = self.getExpertParameterOrDefault("AtmIntegrationDuration", float, 0.576)
        self.tpIntegrationDuration   = self.getExpertParameterOrDefault("tpIntegrationDuration", float, 0.016)
        self.elLimit                 = self.getExpertParameterOrDefault("ElLimit", str, "20 deg")
        self.numTargets              = self.getExpertParameterOrDefault("NumTargets", int, 10)
        self.pointFocusBand          = self.getExpertParameterOrDefault("pointFocusBand", int, 3)
        bandStr                      = self.getExpertParameterOrDefault("bandList", str, "3")
        self.bandList = []
        for s in bandStr.split(','):
            n = int(s)
            if n < 1 or n > 10:
                raise Exception("Invalid band number in band list: '%s'" % s)
            self.bandList.append(n)
        self.logInfo("Band list: %s" % str(self.bandList))

    def generateTunings(self):
        corrType = self._array.getCorrelatorType()
        self._pointFocusSpectralSpec = self._tuningHelper.GenerateSpectralSpec(
                band = self.pointFocusBand,
                intent = "interferometry_continuum",
                corrType = corrType,
                #dualMode = True,
                dualMode = False,
                dump = self.dumpDuration,
                channelAverage = self.channelAverageDuration,
                integration = self.integrationDuration,
                tpSampleTime = self.tpIntegrationDuration)
        self._pointFocusSpectralSpec.name = "Band %d pointing/focus" % self.pointFocusBand
        self._calSpectralSpecs = []
        for band in self.bandList:
            if band == self.pointFocusBand:
                self._calSpectralSpecs.append(self._pointFocusSpectralSpec)
                continue
            ss = self._tuningHelper.GenerateSpectralSpec(
                            band = band,
                            intent = "interferometry_continuum",
                            corrType = corrType,
                            #dualMode = True,
                            dualMode = False,
                            dump = self.dumpDuration,
                            channelAverage = self.channelAverageDuration,
                            integration = self.integrationDuration,
                            tpSampleTime = self.tpIntegrationDuration)
            ss.name = "Band %d pointing" % band
            self._calSpectralSpecs.append(ss)

    def populateSourceList(self):
        self._calSources = self.sourceHelper.getAllSkySources(randomise=True)

    def orderedSpecs(self):
        ret = self._calSpectralSpecs
        if self._reverseSpecs:
            ret = list(reversed(self._calSpectralSpecs))
        self._reverseSpecs = not self._reverseSpecs
        return ret

    def doPointing(self):
        try:
            pointingCal = Observation.PointingCalTarget.PointingCalTarget(self._srcPointFocus, self._pointFocusSpectralSpec)
            pointingCal.setSubscanDuration(self.pointingSubscanDuration)
            pointingCal.setDataOrigin('CHANNEL_AVERAGE_CROSS')
            # For Baseline run avoid extra delay results
            pointingCal.setDelayCalReduction(False)
            if self.alreadyParkedWCAs:
                pointingCal.careAboutWCALOParking = False
            self.logInfo('Executing PointingCal on ' + self._srcPointFocus.sourceName + '...')
            pointingCal.execute(self._obsmode)
            self.logInfo('Completed PointingCal on ' + self._srcPointFocus.sourceName)
            result = pointingCal.checkResult(self._array)
            self.logInfo("Result is: %s" % str(result))
            if len(result) > 0: 
                for key in result.keys():
                    self.logInfo("Found solution for %s using polarization(s) %s" %
                            (key, result[key]))
                pointingCal.applyResult(self._obsmode, result)
            else:
                if not "OSS" in self._array._arrayName:
                    raise Exception("No pointing results!")
        except BaseException, ex:
            print ex
            msg = "Error executing pointing on source %s" % self._srcPointFocus.sourceName
            self.logError(msg)
            self.closeExecution(ex)
            raise ex
        self.alreadyParkedWCAs = (str(self._pointFocusSpectralSpec.FrequencySetup.receiverBand) == "ALMA_RB_03")
                    
    def doFocus(self):
        try:
            focusCal = Observation.FocusCalTarget.FocusCalTarget(
                    SubscanFieldSource = self._srcPointFocus,
                    Axis = 'Z',
                    SpectralSpec = self._pointFocusSpectralSpec,
                    DataOrigin = 'CHANNEL_AVERAGE_CROSS',
                    SubscanDuration = self.focusSubscanDuration,
                    OneWay = False,
                    NumPositions = 7)
            if self.alreadyParkedWCAs:
                focusCal.careAboutWCALOParking = False
            self.logInfo('Executing FocusCal on ' + self._srcPointFocus.sourceName + '...')
            focusCal.execute(self._obsmode)
            self.logInfo('Completed FocusCal on ' + self._srcPointFocus.sourceName)
            result = focusCal.checkResult(self._array)
            self.logInfo("Result is: %s" % str(result))
            if len(result) > 0: 
                for key in result.keys():
                    self.logInfo("Found solution for %s using polarization(s) %s" %
                            (key, result[key]))
                focusCal.applyResult(self._obsmode, result)
            else:
                if not "OSS" in self._array._arrayName:
                    raise Exception("No focus results!")
        except BaseException, ex:
            print ex
            msg = "Error executing focus on source %s" % self._srcPointFocus.sourceName
            self.logError(msg)
            self.closeExecution(ex)
            raise ex
        self.alreadyParkedWCAs = (str(self._pointFocusSpectralSpec.FrequencySetup.receiverBand) == "ALMA_RB_03")
                    
    def doCalSource(self, src):
        for ss in self.orderedSpecs():
            try:
                subscanDuration = self.pointingSubscanDuration
                pointingCal = Observation.PointingCalTarget.PointingCalTarget(src, ss)
                pointingCal.setSubscanDuration(subscanDuration)
                pointingCal.setIntegrationTime(1.0)
                pointingCal.setDataOrigin('CHANNEL_AVERAGE_CROSS')
                pointingCal.setDelayCalReduction(True)
                pointingCal.setCrossOneWay(False)
                if self.alreadyParkedWCAs:
                    pointingCal.careAboutWCALOParking = False
                self.logInfo('Executing PointingCal on ' + src.sourceName + '...')
                pointingCal.execute(self._obsmode)
                self.logInfo('Completed PointingCal on ' + src.sourceName)
            except BaseException, ex:
                print ex
                msg = "Error executing pointing scan on source %s" % src.sourceName
                self.logError(msg)
                self.closeExecution(ex)
                raise ex
            self.alreadyParkedWCAs = (str(ss.FrequencySetup.receiverBand) == "ALMA_RB_03")

    def doCalObservations(self):
        for i in range(self.repeatCount):
            for src in self._calSources:
                if not self._obsmode.isObservable(src, 600):
                    self.logInfo("Skipping source '%s' as not observable" % src.sourceName)
                    continue
                self.doCalSource(src)



Observation.ObsCalBase.checkArray()

survey = ObsCalAllSkyPointing()
survey.parseOptions()
survey.checkAntennas()
survey.startPrepareForExecution()
survey.generateTunings()
survey.populateSourceList()
survey.findPointFocusSource()
survey.completePrepareForExecution()
survey.logInfo("Executing first pointing...")
survey.doPointing()
survey.logInfo("Executing second pointing -- make sure results are good!...")
survey.doPointing()
survey.logInfo("Executing focus...")
survey.doFocus()
survey.logInfo("Executing Calibration observations...")
survey.doCalObservations()
survey.closeExecution()

