#! /usr/bin/env python
#*******************************************************************************
# ALMA - Atacama Large Millimiter Array
# (c) Associated Universities Inc., 2009 
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
#
# "@(#) $Id: ObsCalGrids.py 214236 2015-02-06 05:12:53Z nphillip $"

#
# forcing global imports is due to an OSS problem
#
global sys
import sys
global math
import math
global datetime
import datetime

global CCL
import CCL.Global
import CCL.FieldSource

global Control
import Control

global ControlExceptionsImpl
import ControlExceptionsImpl

global AcsutilPy
import AcsutilPy.FindFile

global Observation
import Observation.AtmCalTarget
import Observation.DelayCalTarget
import Observation.PointingCalTarget
import Observation.FocusCalTarget
import Observation.PhaseCalTarget
import Observation.BandpassCalTarget
import Observation.ObsCalBase


class ObsCalGrids(Observation.ObsCalBase.ObsCalBase):

    azoffset = math.radians(150.0/3600.0)

    def __init__(self):
        Observation.ObsCalBase.ObsCalBase.__init__(self)
        self._srcPointFocus = None
        self._reverseSpecs = False
        self.doneBandpass = False
        self.donePrimaryAmpCal = False

    def parseOptions(self):
        self.repeatCount             = self.getExpertParameterOrDefault("RepeatCount", int, 1)
        self.pointingSubscanDuration = self.getExpertParameterOrDefault("PointingSubscanDuration", float, 5.76)
        self.atmSubscanDuration      = self.getExpertParameterOrDefault("AtmSubscanDuration", float, 5.76)
        self.sbrSubscanDuration      = self.getExpertParameterOrDefault("SBRSubscanDuration", float, 5.76)
        self.focusSubscanDuration    = self.getExpertParameterOrDefault("FocusSubscanDuration", float, 5.76)
        self.dumpDuration            = self.getExpertParameterOrDefault("dumpDuration", float, 0.192)
        self.channelAverageDuration  = self.getExpertParameterOrDefault("channelAverageDuration", float, 0.576)
        self.integrationDuration     = self.getExpertParameterOrDefault("integrationDuration", float, 2.88)
        #self.atmIntegrationDuration  = self.getExpertParameterOrDefault("AtmIntegrationDuration", float, 0.576)
        self.tpIntegrationDuration   = self.getExpertParameterOrDefault("tpIntegrationDuration", float, 0.016)
        self.elLimit                 = self.getExpertParameterOrDefault("ElLimit", str, "20 deg")
        self.HourAngleLimit          = self.getExpertParameterOrDefault("HourAngleLimit", float, 3.0)
        self.numTargets              = self.getExpertParameterOrDefault("NumTargets", int, 10)
        self.pointFocusBand          = self.getExpertParameterOrDefault("pointFocusBand", int, 7)
        bandStr                      = self.getExpertParameterOrDefault("bandList", str, "7,3")
        self.bandList = []
        for s in bandStr.split(','):
            n = int(s)
            if n < 1 or n > 10:
                raise Exception("Invalid band number in band list: '%s'" % s)
            self.bandList.append(n)
        self.logInfo("Band list: %s" % str(self.bandList))

    def generateTunings(self):
        corrType = self._array.getCorrelatorType()
        self._pointFocusSpectralSpec = self._tuningHelper.GenerateSpectralSpec(
                band = self.pointFocusBand,
                intent = "interferometry_continuum",
                corrType = corrType,
                dualMode = False,
                dump = self.dumpDuration,
                channelAverage = self.channelAverageDuration,
                integration = self.integrationDuration,
                tpSampleTime = self.tpIntegrationDuration)
        self._pointFocusSpectralSpec.name = "Band %d pointing/focus" % self.pointFocusBand
        self._calSpectralSpecs = []
        for band in self.bandList:
            ss = self._tuningHelper.GenerateSpectralSpec(
                            band = band,
                            intent = "calsurvey",
                            corrType = corrType,
                            #dualMode = True,
                            dualMode = False,
                            dump = self.dumpDuration,
                            channelAverage = self.channelAverageDuration,
                            integration = self.integrationDuration,
                            tpSampleTime = self.tpIntegrationDuration)
            ss.name = "Band %d calsurvey" % band
            self._calSpectralSpecs.append(ss)

    def populateSourceList(self, nameList=None):
        self._calSources = [self._srcPointFocus]
        self._calSources += self.sourceHelper.getUnobscuredPrimaryFluxCalSources()
        self._calSources += self.sourceHelper.getAllGridSources()
        for i in range(1, len(self._calSources)):
            if self._calSources[i].sourceName == self._srcPointFocus.sourceName:
                self.logInfo("Deleting duplicate source list entry %s" % self._calSources[i].sourceName)
                del self._calSources[i]
                break
        self.logInfo("Source list: %s" % str(self._calSources))

    def orderedSpecs(self):
        ret = self._calSpectralSpecs
        if self._reverseSpecs:
            ret = reversed(self._calSpectralSpecs)
        self._reverseSpecs = not self._reverseSpecs
        return ret

    def doPointing(self):
        try:
            pointingCal = Observation.PointingCalTarget.PointingCalTarget(self._srcPointFocus, self._pointFocusSpectralSpec)
            pointingCal.setSubscanDuration(self.pointingSubscanDuration)
            pointingCal.setDataOrigin('CHANNEL_AVERAGE_CROSS')
            pointingCal.setDelayCalReduction(True)
            self.logInfo('Executing PointingCal on ' + self._srcPointFocus.sourceName + '...')
            pointingCal.execute(self._obsmode)
            self.logInfo('Completed PointingCal on ' + self._srcPointFocus.sourceName)
            result = pointingCal.checkResult(self._array)
            self.logInfo("Result is: %s" % str(result))
            if len(result) > 0: 
                for key in result.keys():
                    self.logInfo("Found solution for %s using polarization(s) %s" %
                            (key, result[key]))
                pointingCal.applyResult(self._obsmode, result)
            else:
                if not "OSS" in self._array._arrayName:
                    raise Exception("No pointing results!")
        except BaseException, ex:
            print ex
            msg = "Error executing pointing on source %s" % self._srcPointFocus.sourceName
            self.logError(msg)
            self.closeExecution(ex)
            raise ex
                    
    def doFocus(self):
        try:
            focusCal = Observation.FocusCalTarget.FocusCalTarget(
                    SubscanFieldSource = self._srcPointFocus,
                    Axis = 'Z',
                    SpectralSpec = self._pointFocusSpectralSpec,
                    DataOrigin = 'CHANNEL_AVERAGE_CROSS',
                    SubscanDuration = self.focusSubscanDuration,
                    OneWay = False,
                    NumPositions = 7)
            self.logInfo('Executing FocusCal on ' + self._srcPointFocus.sourceName + '...')
            focusCal.execute(self._obsmode)
            self.logInfo('Completed FocusCal on ' + self._srcPointFocus.sourceName)
            result = focusCal.checkResult(self._array)
            self.logInfo("Result is: %s" % str(result))
            if len(result) > 0: 
                for key in result.keys():
                    self.logInfo("Found solution for %s using polarization(s) %s" %
                            (key, result[key]))
                focusCal.applyResult(self._obsmode, result)
            else:
                if not "OSS" in self._array._arrayName:
                    raise Exception("No focus results!")
        except BaseException, ex:
            print ex
            msg = "Error executing focus on source %s" % self._srcPointFocus.sourceName
            self.logError(msg)
            self.closeExecution(ex)
            raise ex
                    
    def doCalSource(self, src):
        doneBandpassThisTime = False
        donePrimaryAmpCalThisTime = False
        for ss in self.orderedSpecs():
            try:
                atm = Observation.AtmCalTarget.AtmCalTarget(src, ss, doHotLoad=True)
                atm.setOnlineProcessing(True)
                atm.setDataOrigin('FULL_RESOLUTION_AUTO')
                atm.setDoZero(False)
                atm.setSubscanDuration(self.atmSubscanDuration)
                atm.setIntegrationTime(1.5)
                atm.setWVRCalReduction(True)
                # Applying the results takes a while with lots of
                # antennas, so until we use online WVR, don't bother
                atm.setApplyWVR(False)
                # Currently we need to set a reference source to use an offset.
                atm._referenceSource=atm._source
                try:
                        atm._referenceOffset=CCL.SourceOffset.stroke(self.azoffset,0,0,0,Control.HORIZON)
                        self.logInfo("Using 10.6 and later SourceOffset.stroke for AtmCal reference position")
                except:
                        self.logInfo("Using 10.4 and earlier SourceOffset for AtmCal reference position")
                        atm._referenceOffset=Control.SourceOffset(self.azoffset,0,0,0,Control.HORIZON)
                self.logInfo('Executing AtmCal on ' + src.sourceName + '...')
                atm.execute(self._obsmode)
                self.logInfo('Completed AtmCal on ' + src.sourceName)

                subscanDuration = 0.3e-9 * ss.getMeanFrequency()

                if not self.doneBandpass:
                    ampliCal = Observation.BandpassCalTarget.BandpassCalTarget(src, ss)
                    doneBandpassThisTime = True
                elif not self.donePrimaryAmpCal:
                    ampliCal = Observation.AmplitudeCalTarget.AmplitudeCalTarget(src, ss)
                    donePrimaryAmpCalThisTime = True
                else:
                    ampliCal = Observation.PhaseCalTarget.PhaseCalTarget(src, ss)
                ampliCal.setSubscanDuration(subscanDuration)
                ampliCal.setIntegrationTime(1.0)
                # Tuning is the same as preceding AtmCal, so no need to park WCA LOs again
                ampliCal.careAboutWCALOParking = False
                self.logInfo('Executing AmplitudeCal on ' + src.sourceName + '...')
                ampliCal.execute(self._obsmode)
                self.logInfo('Completed AmplitudeCal on ' + src.sourceName)
                firstSS = False
            except BaseException, ex:
                print ex
                msg = "Error executing cal survey scans on source %s" % src.sourceName
                self.logError(msg)
                self.closeExecution(ex)
                raise ex
        if doneBandpassThisTime:
            self.doneBandpass = True
        if donePrimaryAmpCalThisTime:
            self.donePrimaryAmpCal = True

    def doCalObservations(self):
        for src in self._calSources:
            self.logInfo("Source: %s" % str(src))
            if isinstance(src, CCL.FieldSource.EquatorialSource):
                lstSec = self.sourceHelper.getLSTsecondOfDay()
                RASec = (24*60*60 / 360) * math.degrees(src.sourceCoordinates.longitude.get())
                HA = abs(RASec - lstSec) / 3600.0
                self.logInfo("LST seconds: %d, source RA seconds: %d, hour angle hours: %.3f" % (lstSec, int(RASec), HA))
                if HA > self.HourAngleLimit:
                    self.logInfo("Skipping source '%s' as hour angle over limit of %.3f hours" % (src.sourceName, self.HourAngleLimit))
                    continue
            if not self._obsmode.isObservable(src, 600):
                self.logInfo("Skipping source '%s' as not observable" % src.sourceName)
                continue
            self.doCalSource(src)


Observation.ObsCalBase.checkArray()

survey = ObsCalGrids()
survey.parseOptions()
survey.checkAntennas()
survey.startPrepareForExecution()
survey.generateTunings()
survey.findPointFocusSource()
survey.populateSourceList()
survey.completePrepareForExecution()
survey.logInfo("Executing first pointing...")
survey.doPointing()
survey.logInfo("Executing second pointing -- make sure results are good!...")
survey.doPointing()
survey.logInfo("Executing focus...")
survey.doFocus()
survey.logInfo("Executing third pointing after focus -- make sure results are good!...")
survey.doPointing()
survey.logInfo("Executing Calibration observations...")
survey.doCalObservations()
survey.closeExecution()
