#! /usr/bin/env python
# ALMA - Atacama Large Millimeter Array
# (c) Associated Universities Inc., 2009 - 2011
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that i1t will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
#
# "@(#) $Id: StandardSingleDish.py 212412 2014-12-11 20:28:17Z rmarson $"

#************************************************************************
#   NAME StandardSingleDish.py
#
#   SYNOPSIS This script is inteded to provide basic single dish
#            observations capabilities of single pointing targets.
#
#------------------------------------------------------------------------

#
# forcing global imports is due to an OSS problem
#
global math
import math
global time
from time import time

global Control
import Control

global CCL
from CCL.Global import *
from CCL.ObservingModeBase import SBOutOfTime
import CCL.APDMSchedBlock

global Acspy
import Acspy.Common.TimeHelper

from PyDataModelEnumeration import PyCalibrationDevice

global Observation
import Observation.SchedulingBlock
from Observation.SBExecutionMode import SBExecutionMode, SBExecutionError
from Observation.Global import simulatedArray, addClassToMode, OperatorLogger
from Observation.SBRatioCalTarget import SBRatioCalTarget
from Observation.TargetUtility import *
from Observation.ScanList import ScanList
from Observation.SBExecState import SBExecState


# Ugly hack to get nice logging when we're not written as a class
class StandardSingleDish(Observation.SSRLogger.SSRLogger):
    def __init__(self):
        Observation.SSRLogger.SSRLogger.__init__(self, self.__class__.__name__)
logger = StandardSingleDish()


sb = Observation.SchedulingBlock.SchedulingBlock()

# Get expert parameters as needed here
useScanSequence = bool(int(sb.getExpertParameter('useScanSequence', default=False))) # currently screwed-up with dual mode
sessionControl  = int(sb.getExpertParameter('SessionControl', default=0))
elLimit         = float(sb.getExpertParameter('ElLimit', default=20.))
numFocusCal     = int(sb.getExpertParameter('NumFocusCal', default=1))

#print "####### DEBUG ######"
#print "representativeFrequency %f" % repFrequency
##print "####### DEBUG ######"

logger.logInfo('Session control %d' % sessionControl)
logger.logInfo("useScanSequence value %s" % useScanSequence)

# Get reference to the array and start the exec block
array = getArray()
array.beginExecution()

# Get the Observing Mode and set the maximum execution time
obsmode = array.getInterferometryObservingMode()

# add the SB execution methods
# This is a dangerous way to overwrite the parent class methods
# We should re-think how to do this (PC)
addClassToMode(obsmode, SBExecutionMode)

sbe = SBExecState(array, sb)
#Just start on the sky, to be sure.  Should be in the mode defaults
obsmode.setCalibrationDevice(PyCalibrationDevice.NONE)

obsmode.addSBExecState(sbe)
# Set observing mode restrictions
logger.logInfo("Setting elevation limit to %f degrees" % elLimit)
obsmode.setElevationLimit('%f deg' % elLimit)
endTime = sb.getMaximumExecutionTime()
logger.logInfo("Setting maximum execution time to %s seconds or %s minutes" % (endTime,endTime/60.0))
obsmode.setMaximumSchedBlockTime(endTime)

# Turn off automatic WVR reduction on all targets:
# Turn off automatic WVR reduction on all targets:
# we do not do this in simulation as disableWVRCalibration() is not simulated.
if ( not simulatedArray ):
     obsmode.disableWVRCalibration(True)


t0 = time()
obsmode.adjustDoppler(sb.getTargetList(), sb.getRepresentativeTarget())
logger.logInfo('Doppler adjustment took %2.2f seconds' %(time()-t0))

#
# Turning off the atm cals for pointing targets
# Is this  done underneath?
#for target in sb.getPointingCalTargetList(): target._associatedCalTarget = []
#for target in sb.getFocusCalTargetList(): target._associatedCalTarget = []

#maxPointingSeparation = math.radians(15.0)

#
# Get group list
groupList = sb.getGroupList()
if (len(groupList[1].getTargetList()) > 0):
    targetInfo = obsmode.getScienceTargetInfo(groupList[1])
    logger.logInfo("Science targets are up")
    logger.logInfo("Science targets coordinates %s" % targetInfo['coords'])
    logger.logInfo("Science targets BB frequencies %s" % targetInfo['bbfreq'])
    logger.logInfo("Science targets representative frequencies %s" %\
                    targetInfo['repFreq'])
# Check the visibility of science targets in all groups, (they should need no query expansion)
# Get the coordinates from the science target an pass it for the query
for group in groupList:
    if not obsmode.checkScienceTargetVisibility(group):
        logger.logError('Science targets in group %d are not visible, try later or tomorrow.' \
                        %(groupList.index(group)+1))
        raise SBExecutionError, "No visible science target"



#
# Expand source queries, then associate atm and pointing targets with other targets:
for group in groupList:
    logger.logInfo("Preparing targets for Observing Group %d" % (groupList.index(group)+1))
    #
    #Begin merge
    #
    t0 = time()
    #
    #End merge
    #    
    group.prepareTargets(targetInfo, assocPntToSciTarget = False)
    #
    #Begin merge
    #
    logger.logInfo('Doppler adjustment took %2.2f seconds' %(time()-t0))
    #
    #End merge
    #    


t0 = time()
# We do not need an Atm  calibration for the focus
sb.createAtmCalTargets(targetInfo, True)
sb.createSBRatioCalTargets()
logger.logInfo('Preparation of Atm and SBRatio Cal targets took %2.2f seconds' %(time()-t0))
obsmode.adjustDoppler(sb.getAtmCalTargetList()+sb.getSBRatioCalTargetList(), sb.getRepresentativeTarget())

# Diasable online procesing of SBR, as TelCal's defaults are generally more reliable for a small array of antennas.
logger.logInfo('Disabling online sideband ratio processing')
for target in sb.getSBRatioCalTargetList():
    target.setOnlineProcessing(False)

# be ready for the worst
status = (Control.FAIL, "SB failed")

# This should be the default for real interferometry: we use reference positions only for AtmCals and science targets.
for target in sb.getTargetList():
    target.setUseReferencePosition(target in (sb.getAtmCalTargetList() + sb.getScienceTargetList() ) )

try:
    group = groupList[0]
    groupNumber = groupList.index(group)+1
    logger.logInfo("Start Observing Group %d (Calibrations)" % groupNumber)
    #
    # Pointing now executed in association with the other cal targets.
    if (useScanSequence):
        logger.logInfo("Scan Sequences are enabled")
        scanList = ScanList()
    else:
        scanList = None
        logger.logInfo("Scan Sequences are disabled")

    # Use dual mode always -- this is single dish! 16ms default sampling
    for target in sb.getTargetList():
        ss = target._spectralSpec
        try:
            sqlTest = ss.SquareLawSetup.integrationDuration
            logger.logInfo("SpectralSpec '%s' already has a SquareLawSetup" % str(ss.name))
        except AttributeError:
            ss.SquareLawSetup = CCL.APDMSchedBlock.SquareLawSetup()
            ss.SquareLawSetup.integrationDuration.set(0.016)
            logger.logInfo("SpectralSpec '%s' had no SquareLawSetup, so added one" % str(ss.name))
 
    nFocus, nFocusCleanup = obsmode.executeCalTargetList(group.getFocusCalTargetList(), minimumNumber=numFocusCal)
    if (useScanSequence):
       scanList.execute(obsmode)
    logger.logInfo("Focus Targets: %d executed, %d in Cleanup List" % (nFocus, nFocusCleanup))


    # Now the rest of the groups should contain science targets
    # This won't do scan sequences, testing first
    for group in groupList[1:]:
        while not group.isComplete(obsmode):
            obsmode.executeCalTargetList(group.getDelayCalTargetList(),scanList=scanList)
            if (useScanSequence):
                scanList.execute(obsmode)

            for scienceTarget in group.getScienceTargetList():
                if scienceTarget.isObservable(obsmode):
#                    atmCalTargets=sb.getAtmCalTargetList()
#            atmCalTarget = atmCalTargets[0]
                     assocCalTargets = scienceTarget.getAssociatedCalTarget()
            for target in assocCalTargets :
                target.executeIfNeeded(obsmode)
            scienceTarget.execute(obsmode, assocCalTargets, 450.0)

except SBOutOfTime:
    # Ok there should be just enough time for the cleanup list now
    obsmode.executeCleanupList()

array.endExecution(Control.SUCCESS, "Successful Completion")

#
# ___oOo___
